<?php

/**
 * Cloudrexx
 *
 * @link      http://www.cloudrexx.com
 * @copyright Cloudrexx AG 2007-2015
 *
 * According to our dual licensing model, this program can be used either
 * under the terms of the GNU Affero General Public License, version 3,
 * or under a proprietary license.
 *
 * The texts of the GNU Affero General Public License with an additional
 * permission and of our proprietary license can be found at and
 * in the LICENSE file you have received along with this program.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * "Cloudrexx" is a registered trademark of Cloudrexx AG.
 * The licensing of the program under the AGPLv3 does not imply a
 * trademark license. Therefore any rights, title and interest in
 * our trademarks remain entirely with us.
 */

/**
 * Main controller for Routing
 *
 * @copyright   Cloudrexx AG
 * @author      Project Team SS4U <info@cloudrexx.com>
 * @package     cloudrexx
 * @subpackage  core_routing
 */

namespace Cx\Core\Routing\Controller;

/**
 * Main controller for Routing
 *
 * @copyright   Cloudrexx AG
 * @author      Project Team SS4U <info@cloudrexx.com>
 * @package     cloudrexx
 * @subpackage  core_routing
 */
class ComponentController extends \Cx\Core\Core\Model\Entity\SystemComponentController {
    public function getControllerClasses() {
        // Return an empty array here to let the component handler know that there
        // does not exist a backend, nor a frontend controller of this component.
        return array('Backend', 'EsiWidget');
    }

    /**
     * {@inheritdoc}
     */
    public function getControllersAccessableByJson() {
        return array(
            'EsiWidgetController',
        );
    }

    /**
     * {@inheritdoc}
     */
    public function postInit(\Cx\Core\Core\Controller\Cx $cx)
    {
        $widgetController = $this->getComponent('Widget');
        foreach (
            array(
                'PAGE_ID',
                'PAGE_NODE_ID',
            ) as $widgetName
        ) {
            $widgetController->registerWidget(
                new \Cx\Core_Modules\Widget\Model\Entity\EsiWidget(
                    $this,
                    $widgetName
                )
            );
        }
    }

    public function preResolve(\Cx\Core\Routing\Url $url) {
        if ($this->cx->getMode() != \Cx\Core\Core\Controller\Cx::MODE_FRONTEND) {
            return;
        }

        // apply any existing rewrite rules
        $em = $this->cx->getDb()->getEntityManager();
        $rewriteRuleRepo = $em->getRepository($this->getNamespace() . '\\Model\\Entity\\RewriteRule');
        $rewriteRules = $rewriteRuleRepo->findBy(array(), array('orderNo'=>'asc'));
        $continue = true;
        $originalUrl = clone $url;
        foreach ($rewriteRules as $rewriteRule) {
            try {
                $url = $rewriteRule->resolve($url, $continue);
            } catch (\Exception $e) {
                \DBG::msg('RewriteRule error: '. $rewriteRule->getRegularExpression());
                \DBG::msg($e->getMessage());
                // This is thrown if the regex of the rule is not valid
            }
            if (!$continue) {
                break;
            }
        }

        // abort in case no rewritting has been done
        if ($originalUrl->toString() == $url->toString()) {
            return;
        }

        // detect infinite loops like /foo(/bar)+
        $iterationPoint = strrpos(
            $url->getPath(),
            $originalUrl->getPath()
        );
        // if $iterationPoint is neither 'false' nor '0', then
        // this means that $originalUrl->getPath() is found in
        // $url->getPath() and is a potential infinite loop
        if ($iterationPoint) {
            $redundancy = substr($url->getPath(), 0, $iterationPoint);
            if (substr_count($url->getPath(), $redundancy) > 2) {
                \DBG::msg('Potential infinite loop detected');
                \DBG::msg('Abort resolving');
                \header($_SERVER['SERVER_PROTOCOL'] . ' 502 Bad Gateway');
                // remove CSRF token
                output_reset_rewrite_vars();
                throw new \Cx\Core\Core\Controller\InstanceException();
            }
        }

        // execute external (301/302) redirection (and cache it)
        if (
            $rewriteRule->getRewriteStatusCode() !=
            \Cx\Core\Routing\Model\Entity\RewriteRule::REDIRECTION_TYPE_INTERN
        ) {
            $headers = array(
                'Location' => $url->toString(),
            );
            if ($rewriteRule->getRewriteStatusCode() == 301) {
                array_push(
                    $headers,
                    $_SERVER['SERVER_PROTOCOL'] . ' 301 Moved Permanently'
                );
            }
            $this->getComponent('Cache')->writeCacheFileForRequest(
                null,
                $headers,
                ''
            );
            \Cx\Core\Csrf\Controller\Csrf::header(
                'Location: ' . $url->toString(),
                true,
                intval($rewriteRule->getRewriteStatusCode())
            );
            die();
        }

        // process internal sub-request
        try {
            \DBG::log('Fetching content from ' . str_replace(' ', '%20', $url->toString()));
            $request = new \HTTP_Request2(str_replace(' ', '%20', $url->toString()), \HTTP_Request2::METHOD_GET);
            $request->setConfig(array(
                // disable ssl peer verification
                'ssl_verify_host' => false,
                'ssl_verify_peer' => false,
                // follow HTTP redirect
                'follow_redirects' => true,
                // resend original request to new location
                'strict_redirects' => true,
            ));
            $response = $request->send();
            http_response_code($response->getStatus());
            $content = $response->getBody();
            $headers = array();
            foreach ($response->getHeader() as $key=>$value) {
                if (in_array($key, array(
                    'content-encoding',
                    'transfer-encoding',
                ))) {
                    continue;
                }
                $headers[$key] = $value;
                \Cx\Core\Csrf\Controller\Csrf::header($key . ':' . $value);
            }
            $continue = false;
            $this->getComponent('Cache')->writeCacheFileForRequest(
                null,
                $headers,
                $content
            );
            die($content);
        } catch (\HTTP_Request2_Exception $e) {
            \DBG::dump($e);
        }
    }

    /**
     * Do something after resolving is done
     *
     * USE CAREFULLY, DO NOT DO ANYTHING COSTLY HERE!
     * CALCULATE YOUR STUFF AS LATE AS POSSIBLE
     * @param \Cx\Core\ContentManager\Model\Entity\Page $page       The resolved page
     */
    public function postResolve(\Cx\Core\ContentManager\Model\Entity\Page $page)
    {
        // TODO: The registration of widgets must not be done here.
        //       Instead the registration must be done in postInit hook.

        // Initialize value of FinalStringWidget CANONICAL_LINK as empty
        // string for the case when the requested page does not have a
        // canonical-link.
        $link = '';

        // NOTE: code has been moved to preFinalize()
        return;

        // fetch canonical-link
        $headers = \Env::get('Resolver')->getHeaders();
        if (
            isset($headers['Link']) &&
            preg_match('/^<([^>]+)>;\s+rel="canonical"/', $headers['Link'], $matches)
        ) {
            $canonicalLink = $matches[1];
            
            $link = new \Cx\Core\Html\Model\Entity\HtmlElement('link');
            $link->setAttribute('rel', 'canonical');
            $link->setAttribute('href', $canonicalLink);
        }

        // TODO: Once each componet will have implemented a proper resolve hook
        //       the CANONICAL_LINK widget shall be converted into an EsiWidget.
        $this->getComponent('Widget')->registerWidget(
            new \Cx\Core_Modules\Widget\Model\Entity\FinalStringWidget(
                $this,
                'CANONICAL_LINK',
                (string) $link
            )
        );
    }

    /**
     * Do something before main template gets parsed
     *
     * @param \Cx\Core\Html\Sigma                       $template   The main template
     */
    public function preFinalize(\Cx\Core\Html\Sigma $template) {
        // TODO: note: the registration of the CANONICAL_LINK widget must be moved
        //       to the preInit() hook. Currently, it has to be located in the preFinalize()
        //       hook to make it work. The reason is, that Response::header('Link') is set
        //       in adjustResponse() hook.

        // Initialize value of FinalStringWidget CANONICAL_LINK as empty
        // string for the case when the requested page does not have a
        // canonical-link.
        $link = '';

        // fetch canonical-link
        try {
            $link = $this->getComponent('ContentManager')->fetchAlreadySetCanonicalLink($this->cx->getResponse());
        } catch (\Exception $e) {
            \DBG::msg($e->getMessage());
            // no Link header set -> page doesn't have a canonical-link
        }

        // TODO: Once each componet will have implemented a proper resolve hook
        //       the CANONICAL_LINK widget shall be converted into an EsiWidget.
        $this->getComponent('Widget')->registerWidget(
            new \Cx\Core_Modules\Widget\Model\Entity\FinalStringWidget(
                $this,
                'CANONICAL_LINK',
                (string) $link
            )
        );
    }
}
